#include <Wire.h>

//--------------------------------------Variable definition---------------------------------------------//

//GyroITG3200

float GyroX,GyroY,GyroZ,GyroTemp;
int g_offx = 0;
int g_offy = 0;
int g_offz = 0;
#define ITG3200_Address 0x68

/*-----------ADXL345----------------*/

#define DEVICE (0x53)      //ADXL345 device address
#define TO_READ (6)        //num of bytes we are going to read each time (two bytes for each axis)
byte buff[TO_READ] ;       //6 bytes buffer for saving data read from the device
char str[512];             //string buffer to transform data before sending it to the serial port
int regAddress = 0x32;    //first axis-acceleration-data register on the ADXL345

//*---------Kalman------------------*/

static const double C_0 = 1;
static const double Q_angle=0.001, Q_gyro=0.003, R_angle=0.5, dt=0.005;
double P[2][2] = {{ 1, 0 },{ 0, 1}};
  
double Pdot[4] ={ 0,0,0,0};
double q_bias, angle_err, PCt_0, PCt_1, E, K_0, K_1, t_0, t_1;
double angle, angle_dot;


/*-------------Encoder---------------*/

#define LF 0
#define RT 1

int Lduration,Rduration;
boolean LcoderDir,RcoderDir;
const byte encoder0pinA = 2;
const byte encoder0pinB = 8;
byte encoder0PinALast;
const byte encoder1pinA = 3;
const byte encoder1pinB = 9;
byte encoder1PinALast;

int RotationCoder[2];

/*--------------------------------------*/

float k1,k2,k3,k4; //Adjust the PID Parameters
float acceleration;//Acceleration in the conversion
int theta;
float angle_dot_0;    //Angular velocity
float y_value;//y value of acceleration sensor

int turn_flag=0;

int pwm;
int pwm_R,pwm_L;
float range;
float range_error_all;
float wheel_speed;
float last_wheel;
float error_a=0;

/*-----Pin-----*/

int AD0pin=0;
int AD1pin=1;
int AD2pin=2;
int AD3pin=3;

int E1 = 5;     //M1 Speed Control 
int E2 = 6;     //M2 Speed Control 
int M1 = 4;    //M1 Direction Control 
int M2 = 7;    //M1 Direction Control   

//------------------------------------------End of definition----------------------------------------------//

void setup()
{
  Wire.begin();           // join i2c bus (address optional for master)
  Serial.begin(9600);    // start serial 

  //Turning on the ADXL345
  writeTo(DEVICE, 0x2D, 0);      
  writeTo(DEVICE, 0x2D, 16);
  writeTo(DEVICE, 0x2D, 8);
  //Turning on the ITG3200
  delay(100);
  initGyro();
  delay(100);
  GyroCalibrate();
  delay(100);

  pinMode(4, OUTPUT);
  pinMode(5, OUTPUT);
  pinMode(6, OUTPUT);
  pinMode(7, OUTPUT);
  
  //The balance PID Parameters which read from The ADPin0-3
  k1=24.80;
  k2=9.66;
  k3=4.14;
  k4=0.99;
  
  CoderInit();//Initialize the module
}

// **************************
// Init the Incoder
// **************************

void CoderInit()
{
  LcoderDir = true;
  RcoderDir = true;	
  pinMode(8,INPUT);
  pinMode(9,INPUT);
  attachInterrupt(LF, LwheelSpeed, CHANGE);
  attachInterrupt(RT, RwheelSpeed, CHANGE);
}

// **************************
// I2C Gyroscope ITG3200 
// **************************
void initGyro() {
   Wire.beginTransmission(ITG3200_Address); 
   Wire.write(0x3E);  
   Wire.write(0x00);   
   Wire.endTransmission(); 
   
   Wire.beginTransmission(ITG3200_Address); 
   Wire.write(0x15);  
   Wire.write(0x07);   
   Wire.endTransmission(); 
   
   Wire.beginTransmission(ITG3200_Address); 
   Wire.write(0x16);  
   Wire.write(0x1E);   // +/- 2000 dgrs/sec, 1KHz, 1E, 19
   Wire.endTransmission(); 
   
   Wire.beginTransmission(ITG3200_Address); 
   Wire.write(0x17);  
   Wire.write(0x00);   
   Wire.endTransmission(); 
   
   
}

void GyroCalibrate(){

     int tmpx = 0;
     int tmpy = 0;
     int tmpz = 0; 

     g_offx = 0;
     g_offy = 0;
     g_offz = 0;
 
      for (char i = 0;i<10;i++)
       {
         delay(10);  
         GyroRead();
         tmpx += GyroX;
         tmpy += GyroY;
         tmpz += GyroZ; 
       }  
      g_offx = tmpx/10;
      g_offy = tmpy/10;
      g_offz = tmpz/10;
    
}


void GyroRead() {
  Wire.beginTransmission(ITG3200_Address); 
  Wire.write(0x1B);       
  Wire.endTransmission(); 
  
  Wire.beginTransmission(ITG3200_Address); 
  Wire.requestFrom(ITG3200_Address, 8);    // request 8 bytes from ITG3200
  
  int i = 0;
  byte buff[8];
  while(Wire.available())    
  { 
    buff[i] = Wire.read(); 
    i++;
  }
  Wire.endTransmission(); 
    
  GyroX = ((buff[4] << 8) | buff[5]) - g_offx;
  GyroY = ((buff[2] << 8) | buff[3]) - g_offy;
  GyroZ = ((buff[6] << 8) | buff[7]) - g_offz;
  GyroTemp = (buff[0] << 8) | buff[1]; // temperature 
 
  
}

// *******************************
// Read the Speed from the Encoder
// *******************************

void LwheelSpeed()
{
  int Lstate = digitalRead(encoder0pinA);
  if((encoder0PinALast == LOW)&&Lstate==HIGH)
  {
    int val = digitalRead(encoder0pinB);
    if(val == LOW && LcoderDir)  LcoderDir = false; //Lreverse
    else if(val == HIGH && !LcoderDir)  LcoderDir = true;  //Lforward
  }
  encoder0PinALast = Lstate;

  if(!LcoderDir)  Lduration++;
  else  Lduration--;

  if(true)  RotationCoder[LF] ++;
  else RotationCoder[LF] = 0;
}


void RwheelSpeed()
{
  int Rstate = digitalRead(encoder1pinA);
  if((encoder1PinALast == LOW)&&Rstate==HIGH)
  {
    int val = digitalRead(encoder1pinB);
    if(val == LOW && RcoderDir)  RcoderDir = false; //Rreverse
    else if(val == HIGH && !RcoderDir)  RcoderDir = true;  //Rforward
  }
  encoder1PinALast = Rstate;

  if(!RcoderDir)  Rduration--;
  else  Rduration++;

  if(true)  RotationCoder[RT] ++;
  else RotationCoder[RT] = 0;
}

// **************************
// Kalman Filter
// **************************

void Kalman_Filter(double angle_m,double gyro_m)   
{
  angle+=(gyro_m-q_bias) * dt;
  angle_err = angle_m - angle;
  Pdot[0]=Q_angle - P[0][1] - P[1][0];
  Pdot[1]=- P[1][1];
  Pdot[2]=- P[1][1];
  Pdot[3]=Q_gyro;
  P[0][0] += Pdot[0] * dt;
  P[0][1] += Pdot[1] * dt;
  P[1][0] += Pdot[2] * dt;
  P[1][1] += Pdot[3] * dt;
  PCt_0 = C_0 * P[0][0];
  PCt_1 = C_0 * P[1][0];
  E = R_angle + C_0 * PCt_0;
  K_0 = PCt_0 / E;
  K_1 = PCt_1 / E;
  t_0 = PCt_0;
  t_1 = C_0 * P[0][1];
  P[0][0] -= K_0 * t_0;
  P[0][1] -= K_0 * t_1;
  P[1][0] -= K_1 * t_0;
  P[1][1] -= K_1 * t_1;
  angle	+= K_0 * angle_err;  //Optimal angle
  q_bias += K_1 * angle_err;
  angle_dot = gyro_m-q_bias; //Optimal angular velocity
}

// **************************
// Calculate the Angle
// **************************

void Angle_calculate(void)
{   
  readFrom(DEVICE, regAddress, TO_READ, buff); //read the acceleration data from the ADXL345
  y_value = (((int)buff[3])<< 8) | buff[2]; 
  acceleration=y_value/267;// current acceleration
  theta=asin(acceleration/0.98)*57.3-error_a;
  
   GyroRead();
  angle_dot_0=GyroY / 14.375;       //角速度换算
  Kalman_Filter(theta,angle_dot_0); //卡尔曼滤波

}
// **************************
// Calculate the pwm
// **************************
void pwm_calculate()
{
  int range_error;

  range+=(Lduration+Rduration)*0.5;
  range*=0.9;
  range_error=Lduration-Rduration;
  range_error_all+=range_error;

  wheel_speed=range-last_wheel;
  last_wheel=range;
  
  pwm=angle*k1+angle_dot*k2+range*k3+wheel_speed*k4;    //use PID tho calculate the pwm
 
  if(pwm>255)                               //Maximum Minimum Limitations
    pwm=255;
  if(pwm<-255)
    pwm=-255;
    

    if(turn_flag==0)
    {
     pwm_R=pwm+range_error_all;
     pwm_L=pwm-range_error_all;
    }
    else
    {
        pwm_R=pwm-turn_flag*68;  //Direction PID control
        pwm_L=pwm+turn_flag*68;
        range_error_all=0;     //clean
    }
    
     Lduration = 0;//clean
     Rduration = 0;
   
 
}
// **************************
// Set the motor
// **************************

void set_motor()
{
   if(pwm_R>0)   //Set left wheel's direction and speed
  {
     digitalWrite(M1,HIGH);
     analogWrite(E1,pwm_R);
  }
  else if(pwm_R<0)
  {
     digitalWrite(M1,LOW);
     analogWrite(E1,-pwm_R);
  }
   if(pwm_L>0)   //Set Right wheel's direction and speed
  {
     digitalWrite(M2,HIGH);
     analogWrite(E2,pwm_L);
  }
  else if(pwm_L<0)
  {
     digitalWrite(M2,LOW);
     analogWrite(E2,-pwm_L);
  }
}

//---------------- I2C Functions --------------------------------------------------------------
//Writes val to address register on device
void writeTo(int device, byte address, byte val) {
  Wire.beginTransmission(device); //start transmission to device 
  Wire.write(address);        // send register address
  Wire.write(val);        // send value to write
  Wire.endTransmission(); //end transmission
}

//reads num bytes starting from address register on device in to buff array
void readFrom(int device, byte address, int num, byte buff[])
{
  Wire.beginTransmission(device); //start transmission to device 
  Wire.write(address);        //sends address to read from
  Wire.endTransmission(); //end transmission

    Wire.beginTransmission(device); //start transmission to device
  Wire.requestFrom(device, num);    // request 6 bytes from device

  int i = 0;
  while(Wire.available())    //device may send less than requested (abnormal)
  { 
    buff[i] = Wire.read(); // receive a byte
    i++;
  }
  Wire.endTransmission(); //end transmission
}

//-----------------------I2C Functions end---------------------------------------------------------

// **************************
// Remote control code
// **************************
void control()
{
  if(Serial.available()){
    int val;
    val=Serial.read();
    switch (val)
    {
    case 'a':   //forward
         if(error_a>-8)
         
          { error_a-=0.1;  }
           break;
    case 'b':   //back
          if(error_a<8)
          { error_a+=0.1;  }
           break;
    case 's':   //stop
      error_a=0;
      turn_flag=0;
      break;
    case 'r':  //right
      turn_flag=-1;
      break;
    case 'l':  //left
      turn_flag=1;
      break;
      
    default:
      break;
      
    }
  }
}


void loop()
{
  // k1=analogRead(AD0pin)*0.05;
  // k2=analogRead(AD1pin)*0.03;
  // k3=analogRead(AD2pin)*0.03;
  // k4=analogRead(AD3pin)*0.03;
  control();
  Angle_calculate();
  pwm_calculate();
  set_motor();
  delay(2);
  
 //Read the ADPin0-3
 /* Serial.print(k1);
   Serial.print("_");
     Serial.print(k2);
       Serial.print("_");
        Serial.print(k3);
         Serial.print("_");
          Serial.println(k4);
          delay(1000);*/
    
}
